/*
 *  fs/partitions/mac.c
 *
 *  Code extracted from drivers/block/genhd.c
 *  Copyright (C) 1991-1998  Linus Torvalds
 *  Re-organised Feb 1998 Russell King
 */

#include <linux/config.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/blk.h>
#include <linux/ctype.h>

#include <asm/system.h>

#include "check.h"
#include "mac.h"

#ifdef CONFIG_PPC
extern void note_bootable_part(kdev_t dev, int part);
#endif

/*
 * Code to understand MacOS partition tables.
 */

static inline void mac_fix_string(char *stg, int len)
{
	int i;

	for (i = len - 1; i >= 0 && stg[i] == ' '; i--)
		stg[i] = 0;
}

// #define mbe16_to_cpu(x) (x)
// #define mbe32_to_cpu(x) (x)
#define mbe16_to_cpu(x) be16_to_cpu(x)
#define mbe32_to_cpu(x) be32_to_cpu(x)

int mac_partition(struct gendisk *hd, kdev_t dev, unsigned long fsec, int first_part_minor)
{
	struct buffer_head *bh;
	int blk, blocks_in_map, cblk, b;
	int tivo_only = 0;
//	int dev_bsize, dev_pos, pos;
	unsigned secsize;
// #ifdef CONFIG_PPC
//	int found_root = 0;
//	int found_root_goodness = 0;
// #endif
	struct mac_partition *part;
	struct mac_driver_desc *md;

//	dev_bsize = get_ptable_blocksize(dev);
//	dev_pos = 0;
	/* Get 0th block and look at the first partition map entry. */
	if ((bh = bread(dev, 0, 1024)) == 0) {
	    printk("%s: error reading partition table\n",
		   kdevname(dev));
	    return -1;
	}
	cblk = 0;
	md = (struct mac_driver_desc *) bh->b_data;
        printk("Signature %x, be16 Signature %x\n",md->signature,
                mbe16_to_cpu(md->signature));
        switch (mbe16_to_cpu(md->signature)) {
                case TIVO_BOOT_MAGIC:
                        tivo_only = 0;
                        secsize = 512;
                        part = (struct mac_partition *) (bh->b_data + secsize);
                        break;
                default:
                        brelse(bh);
                        printk("%s: block 0 has signature %x rather than %x\n",
                                kdevname(dev), mbe16_to_cpu(md->signature),
                                TIVO_BOOT_MAGIC);
                        return 0;
        }
       if (mbe16_to_cpu(part->signature) != MAC_PARTITION_MAGIC) {
               brelse(bh);
               printk("%s: first partition has signature %x rather %x\n",
                       kdevname(dev), be16_to_cpu(part->signature), MAC_PARTITION_MAGIC);
               return 0;               /* not a MacOS disk */
       }
       blocks_in_map = mbe32_to_cpu(part->map_count);
       printk("Blocks in Map = %x\n", blocks_in_map);
       for (blk = 1; blk <= blocks_in_map; ++blk) {
               b = (blk * secsize) / 1024;
               if (b != cblk) {
                        brelse(bh);
                        cblk = b;
                        if ((bh = bread(dev, cblk, 1024)) == 0) {
                            printk("%s: error reading partition table\n",
                            kdevname(dev));
                            return -1;
                        }
               }
               part = (struct mac_partition *)
                       (bh->b_data + (blk * secsize) % 1024);
               if (mbe16_to_cpu(part->signature) != MAC_PARTITION_MAGIC) {
                       printk("Block %u has sig %x rather than %x\n",
                               blk, mbe16_to_cpu(part->signature), MAC_PARTITION_MAGIC);
                       break;
               }
               blocks_in_map = mbe32_to_cpu(part->map_count);

               printk(" mac st=%x sz=%x name='%s' t='%s' bim=%x\n",
                       mbe32_to_cpu(part->start_block),
                       mbe32_to_cpu(part->block_count),
                       part->name, part->type, blocks_in_map);

               if (tivo_only && strncmp(part->type, "TiVo", 4) != 0) {
                       continue;
               }
                
		add_gd_partition(hd, first_part_minor,
			mbe32_to_cpu(part->start_block) * (secsize/512),
			mbe32_to_cpu(part->block_count) * (secsize/512));

#ifdef CONFIG_PPC
		/*
		 * If this is the first bootable partition, tell the
		 * setup code, in case it wants to make this the root.
		 */
		if (_machine == _MACH_Pmac) {
			int goodness = 0;

			mac_fix_string(part->processor, 16);
			mac_fix_string(part->name, 32);
			mac_fix_string(part->type, 32);					
		    
			if ((be32_to_cpu(part->status) & MAC_STATUS_BOOTABLE)
			    && strcasecmp(part->processor, "powerpc") == 0)
				goodness++;

			if (strcasecmp(part->type, "Apple_UNIX_SVR2") == 0
			    || strcasecmp(part->type, "Linux_PPC") == 0) {
				int i, l;

				goodness++;
				l = strlen(part->name);
				if (strcmp(part->name, "/") == 0)
					goodness++;
				for (i = 0; i <= l - 4; ++i) {
					if (strnicmp(part->name + i, "root",
						     4) == 0) {
						goodness += 2;
						break;
					}
				}
				if (strnicmp(part->name, "swap", 4) == 0)
					goodness--;
			}

			if (goodness > found_root_goodness) {
				found_root = blk;
				found_root_goodness = goodness;
			}
		}
#endif /* CONFIG_PPC */

		++first_part_minor;
	}
#ifdef CONFIG_PPC
	if (found_root_goodness)
		note_bootable_part(dev, found_root);
#endif
	brelse(bh);
	printk("\n");
	return 1;
}

